/*
 *    file: CGContext.cpp
 *  author: Bastian Spiegel <bs@tkscript.de>
 *
 * changed: 21Nov2007, 26Dec2007
 *
 * license: MIT
 *
Copyright (c) 2007-2008 Bastian Spiegel, Carsten Busse

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

#include "tkcg.h"

_CGContext::_CGContext(void) {
   cg_context       = NULL;
   fragment_profile = NULL;
   vertex_profile   = NULL;
}

_CGContext::~_CGContext() {
   destroy();

   // TODO: Add refcounting..
   fragment_profile = NULL;
   vertex_profile   = NULL;
}

_CGhandle _CGContext::getCgHandle(void) {
   return (_CGhandle) cg_context;
}

void _CGContext::setCgHandle(_CGhandle _cgHandle) {
   cg_context = (CGcontext) _cgHandle;
   b_destroy = 0;
}

void _CGContext::create(void) {
   destroy();
   
   cg_context = cgCreateContext();
   if(Dhandlecgexception())
   {
      om_CGContext.addObject(this);
      b_destroy = 1;
   }
}

void _CGContext::destroy(void) {
   if(NULL != cg_context)
   {
      om_CGContext.removeObject(this);
      if(b_destroy)
      {
         cgDestroyContext(cg_context);
      }
      cg_context = NULL;
      b_destroy = 0;
   }
}

void _CGContext::setFragmentProfile(YAC_Object *_profile) {
   if(YAC_CHK(_profile, clid_CGProfile))
   {
      fragment_profile = (_CGProfile *) _profile;
   }
   else
   {
      Draiseinvalidpointerexception("Invalid fragment profile handle.");
   }
}

void _CGContext::setVertexProfile(YAC_Object *_profile) {
   if(YAC_CHK(_profile, clid_CGProfile))
   {
      vertex_profile = (_CGProfile *) _profile;
   }
   else
   {
      Draiseinvalidpointerexception("Invalid vertex profile handle.");
   }
}

void _CGContext::createVertexProgramFromSource(YAC_Object *_source, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r) {
   createProgramFromString(_source, _entry, _args, _r, CG_SOURCE, vertex_profile);
}

void _CGContext::createVertexProgramFromLocalSource(YAC_Object *_fileName, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r) {
   createProgramFromLocalFile(_fileName, _entry, _args, _r, CG_SOURCE, vertex_profile);
}

void _CGContext::createVertexProgramFromLocalObject(YAC_Object *_fileName, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r) {
   createProgramFromLocalFile(_fileName, _entry, _args, _r, CG_OBJECT, vertex_profile);
}

void _CGContext::createFragmentProgramFromSource(YAC_Object *_source, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r) {
   createProgramFromString(_source, _entry, _args, _r, CG_SOURCE, fragment_profile);
}

void _CGContext::createFragmentProgramFromLocalSource(YAC_Object *_fileName, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r) {
   createProgramFromLocalFile(_fileName, _entry, _args, _r, CG_SOURCE, fragment_profile);
}

void _CGContext::createFragmentProgramFromLocalObject(YAC_Object *_fileName, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r) {
   createProgramFromLocalFile(_fileName, _entry, _args, _r, CG_OBJECT, fragment_profile);
}


void _CGContext::createProgramFromLocalFile(YAC_Object *_fileName, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r, 
                                            CGenum _programType, _CGProfile *_profile) 
{
   _r->initNull();

   if(YAC_VALID(_profile))
   {
      if(YAC_BCHK(_fileName, YAC_CLID_STRING))
      {
         const char *fileName = (char*) ((YAC_String*)_fileName)->chars;
         const char *entry = NULL;
         const char *args[32];
         const char **argsp;

         if(YAC_BCHK(_entry, YAC_CLID_STRING))
         {
            entry = (char*) ((YAC_String*)_entry)->chars;
         }

         if(YAC_VALID(_args))
         {
            sUI numArgs = _args->yacArrayGetNumElements();
            if(numArgs > 0)
            {
               if(numArgs > 31)
               {
                  yac_host->printf("[~~~] CGContext::createProgramFromLocalFile: warning: truncating args array to 31 elements.\n");
                  numArgs = 31;
               }
               sUI i;
               YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
               for(i=0; i<numArgs; i++)
               {
                  YAC_Value r;
                  _args->yacArrayGet(ctx, i, &r);
                  if(r.type == YAC_TYPE_STRING)
                  {
                     args[i] = (char*) r.value.string_val->chars;
                  }
                  else
                  {
                     Draisetypemismatchexception("Program args element is not a String.");
                     return;
                  }
               }
               args[i] = NULL;
               argsp = args;
            }
         }
         else
         {
            argsp = NULL;
         }

         CGprogram cgProgram = cgCreateProgramFromFile(cg_context, _programType, fileName, _profile->cg_profile, entry, argsp);

         if(Dhandlecgexception())
         {
            _CGProgram *p = (_CGProgram *) yac_host->yacNewByID(clid_CGProgram);
            if(NULL != p)
            {
               p->init(cgProgram);
               _r->initObject(p, 1);
               om_CGProgram.addObject(p);
            }
            else
            {
               cgDestroyProgram(cgProgram);
               Draiseinvalidpointerexception("Failed to create CGProgram object.");
            }
         }
      
      }
      else
      {
         Draiseinvalidpointerexception("Invalid <fileName> object.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Current vertex profile object is invalid.");
   }
}

void _CGContext::createProgramFromString(YAC_Object *_source, YAC_Object *_entry, YAC_Object *_args, YAC_Value *_r, 
                                         CGenum _programType, _CGProfile *_profile) 
{
   _r->initNull();

   if(YAC_VALID(_profile))
   {
      if(YAC_BCHK(_source, YAC_CLID_STRING))
      {
         const char *source = (char*) ((YAC_String*)_source)->chars;
         const char *entry = NULL;
         const char *args[32];
         const char **argsp;

         if(YAC_BCHK(_entry, YAC_CLID_STRING))
         {
            entry = (char*) ((YAC_String*)_entry)->chars;
         }

         if(YAC_VALID(_args))
         {
            sUI numArgs = _args->yacArrayGetNumElements();
            if(numArgs > 0)
            {
               if(numArgs > 31)
               {
                  yac_host->printf("[~~~] CGContext::createProgramFromString: warning: truncating args array to 31 elements.\n");
                  numArgs = 31;
               }
               sUI i;
               YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
               for(i=0; i<numArgs; i++)
               {
                  YAC_Value r;
                  _args->yacArrayGet(ctx, i, &r);
                  if(r.type == YAC_TYPE_STRING)
                  {
                     args[i] = (char*) r.value.string_val->chars;
                  }
                  else
                  {
                     Draisetypemismatchexception("Program args element is not a String.");
                     return;
                  }
               }
               args[i] = NULL;
               argsp = args;
            }
         }
         else
         {
            argsp = NULL;
         }

         CGprogram cgProgram = cgCreateProgram(cg_context, _programType, source, _profile->cg_profile, entry, argsp);

         if(Dhandlecgexception())
         {
            _CGProgram *p = (_CGProgram *) yac_host->yacNewByID(clid_CGProgram);
            if(NULL != p)
            {
               p->init(cgProgram);
               _r->initObject(p, 1);
               om_CGProgram.addObject(p);
            }
            else
            {
               cgDestroyProgram(cgProgram);
               Draiseinvalidpointerexception("Failed to create CGProgram object.");
            }
         }
      
      }
      else
      {
         Draiseinvalidpointerexception("Invalid <fileName> object.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Current vertex profile object is invalid.");
   }
}

sSI _CGContext::getManageTextureParameters(void) {
   CGbool b = cgGLGetManageTextureParameters(cg_context);
   Dhandlecgexception();
   return (CG_TRUE==b) ? 1 : 0;
}

void _CGContext::setManageTextureParameters(sSI _b) {
   CGbool b = _b ? CG_TRUE : CG_FALSE;
   cgGLSetManageTextureParameters(cg_context, b);
   Dhandlecgexception();
}

void _CGContext::registerStates(void) {
   cgGLRegisterStates(cg_context);
   Dhandlecgexception();
}

void _CGContext::createEffectFromLocalFile(YAC_Object *_fileName, YAC_Object *_args, YAC_Value *_r)
{
   _r->initNull();

   if(YAC_BCHK(_fileName, YAC_CLID_STRING))
   {
      const char *fileName = (char*) ((YAC_String*)_fileName)->chars;
      const char *args[32];
      const char **argsp;
      
      if(YAC_VALID(_args))
      {
         sUI numArgs = _args->yacArrayGetNumElements();
         if(numArgs > 0)
         {
            if(numArgs > 31)
            {
               yac_host->printf("[~~~] CGContext::createEffectFromLocalSource: warning: truncating args array to 31 elements.\n");
               numArgs = 31;
            }
            sUI i;
            YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
            for(i=0; i<numArgs; i++)
            {
               YAC_Value r;
               _args->yacArrayGet(ctx, i, &r);
               if(r.type == YAC_TYPE_STRING)
               {
                  args[i] = (char*) r.value.string_val->chars;
               }
               else
               {
                  Draisetypemismatchexception("Effect args element is not a String.");
                  return;
               }
            }
            args[i] = NULL;
            argsp = args;
         }
      }
      else
      {
         argsp = NULL;
      }
      
      CGeffect cgEffect = cgCreateEffectFromFile(cg_context, fileName, argsp);
      
      if(Dhandlecgexception())
      {
         _CGEffect *e = (_CGEffect *) yac_host->yacNewByID(clid_CGEffect);
         if(NULL != e)
         {
            e->init(cgEffect);
            _r->initObject(e, 1);
            om_CGEffect.addObject(e);
         }
         else
         {
            cgDestroyEffect(cgEffect);
            Draiseinvalidpointerexception("Failed to create CGEffect object.");
         }
      }
      
   }
   else
   {
      Draiseinvalidpointerexception("Invalid <fileName> object.");
   }
}

void _CGContext::createParameter(sSI _cgType, YAC_Value *_r) {
   _r->initNull();

   CGparameter cgParameter = cgCreateParameter(cg_context, (CGtype) _cgType);
   if(Dhandlecgexception())
   {
      _CGParameter *p = (_CGParameter *) yac_host->yacNewByID(clid_CGParameter);
      if(NULL != p)
      {
         p->init(cgParameter);
         _r->initObject(p, 1);
         om_CGParameter.addObject(p);
      }
      else
      {
         cgDestroyParameter(cgParameter);
         Draiseinvalidpointerexception("Failed to create CGParameter object.");
      }
   }
}

void _CGContext::setAutoCompile(sSI _autoCompileMode) {
   cgSetAutoCompile(cg_context, (CGenum) _autoCompileMode);
   Dhandlecgexception();
}

void _CGContext::getLastListing(YAC_Value *_r) {

   _r->initString(NULL, 0);

   const char *listing = cgGetLastListing(cg_context);
   if(Dhandlecgexception())
   {
      YAC_String *s = (YAC_String *) yac_host->yacNewByID(YAC_CLID_STRING);
      if(NULL != s)
      {
         s->copy((char*)listing);
         _r->initString(s, 1);
      }
      else
      {
         Draiseinvalidpointerexception("Failed to create String object.");
      }
   }
   
}

void _CGContext::createState(YAC_Object *_name, sSI _cgType, YAC_Value *_r) {
   _r->initNull();

   if(YAC_BCHK(_name, YAC_CLID_STRING))
   {
      CGstate cgState = cgCreateState(cg_context, (char*) ((YAC_String*)_name)->chars, (CGtype) _cgType);
      if(Dhandlecgexception())
      {
         _CGState *s = (_CGState*) yac_host->yacNewByID(clid_CGState);
         if(NULL != s)
         {
            s->init(cgState);
            _r->initObject(s, 1);
            om_CGState.addObject(s);
         }
         else
         {
            Draiseinvalidpointerexception("Failed to create CGState object.");
         }
      }
   }
}
