/*
 *    file: CGEffect.cpp
 *  author: Bastian Spiegel <bs@tkscript.de>
 *
 * changed: 21Nov2007, 26Dec2007
 *
 * license: MIT
 *
Copyright (c) 2007-2008 Bastian Spiegel, Carsten Busse

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

#include "tkcg.h"


_CGEffect::_CGEffect(void) {
   cg_effect = NULL;
}

_CGEffect::~_CGEffect() {
   destroy();
}


_CGhandle _CGEffect::getCgHandle(void) {
   return (_CGhandle) cg_effect;
}

void _CGEffect::setCgHandle(_CGhandle _cgHandle) {
   cg_effect = (CGeffect) _cgHandle;
   b_destroy = 0;
}

void _CGEffect::destroy(void) {
   if(NULL != cg_effect)
   {
      om_CGEffect.removeObject(this);
      if(b_destroy)
      {
         cgDestroyEffect(cg_effect);
         Dhandlecgexception();
         b_destroy = 0;
      }
      cg_effect = NULL;
   }
}

void _CGEffect::init(CGeffect _cgEffect) {
   destroy();
   cg_effect = _cgEffect;
   b_destroy = 1;
}

void _CGEffect::createParameter(YAC_String *_name, sSI _cgType, YAC_Value *_r) {
   _r->initObject(NULL, 0);

   if(YAC_BCHK(_name, YAC_CLID_STRING))
   {
      const char *name = (char*) ((YAC_String*)_name)->chars;
      CGparameter cgParameter = cgCreateEffectParameter(cg_effect, name, (CGtype) _cgType);
      if(Dhandlecgexception())
      {
         _CGParameter *p = (_CGParameter *) yac_host->yacNewByID(clid_CGParameter);
         if(NULL != p)
         {
            p->init(cgParameter);
            _r->initObject(p, 1);
            om_CGParameter.addObject(p);
         }
         else
         {
            cgDestroyParameter(cgParameter);
            Draiseinvalidpointerexception("Failed to create CGParameter object.");
         }
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid <name> string object.");
   }
}

void _CGEffect::createFloatParameter(YAC_String *_name, YAC_Value *_r) {
   createParameter(_name, CG_FLOAT, _r);
}

void _CGEffect::getName(YAC_Value *_r) {

   _r->initString(NULL, 0);

   const char *name = cgGetEffectName(cg_effect);
   if(Dhandlecgexception())
   {
      YAC_String *s = (YAC_String *) yac_host->yacNewByID(YAC_CLID_STRING);
      if(NULL != s)
      {
         s->copy((char*) name);
         _r->initString(s, 1);
      }
      else
      {
         Draiseinvalidpointerexception("Failed to create String object.");
      }
   }
   
}

void _CGEffect::getContext(YAC_Value *_r) {
   _r->initNull();
   CGcontext cgContext = cgGetEffectContext(cg_effect);
   if(Dhandlecgexception())
   {
      om_CGContext.findOrCreateByHandle(cgContext, clid_CGContext, _r);
   }
}
