/*
 *    file: CGObjectMap.cpp
 *  author: Bastian Spiegel <bs@tkscript.de>
 *
 * changed: 21Nov2007, 26Dec2007
 *
 * license: MIT
 *
Copyright (c) 2007-2008 Bastian Spiegel, Carsten Busse

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

#include "tkcg.h"


CGObjectMap::CGObjectMap(void) {
   first = NULL;
}

CGObjectMap::~CGObjectMap() {

   deleteEntries();
}

void CGObjectMap::deleteEntries(void) {
   CGObjectMapEntry *c = first;

   while(NULL != c)
   {
      CGObjectMapEntry *n = c->next;
      
      delete c;
      c = n;
   }

   first = NULL;
}

void CGObjectMap::findOrCreateByHandle(_CGhandle _handle, sUI _clid, YAC_Value *_r) {
   CGObjectMapEntry *c = first;
   while(c != NULL)
   {
      if(c->handle == _handle)
      {
         _r->initObject(c->object, 0);
         yac_host->printf("xxx CGObjectMap::found object %p for handle %p.\n", c->object, _handle);
         return;
      }
      c = c->next;
   }
   // Not found, create new object
   _CGObject *o = (_CGObject*) yac_host->yacNewByID(_clid);
   if(NULL != o)
   {
      o->setCgHandle(_handle);
      _r->initObject(o, 1);

      yac_host->printf("xxx CGObjectMap: creating new entry for object %p for handle %p.\n", o, _handle);

      CGObjectMapEntry *e = new CGObjectMapEntry();
      e->next = NULL;
      e->object = o;
      e->handle = _handle;
      c = first;
      if(NULL == c)
      {
         first = e;
      }
      else
      {
         while(NULL != c->next)
         {
            c = c->next;
         }
         c->next = e;
      }
   }
   else
   {
      _r->initNull();
      Draisecgexception("Failed to create new CG object.");
   }
}

void CGObjectMap::addObject(_CGObject *_object) {
   CGObjectMapEntry *c = first;
   CGObjectMapEntry *p = NULL;
   while(c != NULL)
   {
      p = c;
      c = c->next;
   }
   CGObjectMapEntry *e = new CGObjectMapEntry();
   e->next = NULL;
   e->object = _object;
   e->handle = _object->getCgHandle();

   yac_host->printf("xxx CGObjectMap: adding object %p with handle %p to map.\n", _object, e->handle);

   if(NULL != p)
   {
      p->next = e;
   }
   else
   {
      first = e;
   }
}

void CGObjectMap::removeObject(_CGObject *_object) {
   CGObjectMapEntry *c = first;
   CGObjectMapEntry *p = NULL;
   while(NULL != c)
   {
      if(c->object == _object)
      {
         CGObjectMapEntry *n = c->next;
         if(NULL == p)
         {
            first = n;
         }
         else
         {
            p->next = n;
         }
         delete c;
         return;
      }
      p = c;
      c = c->next;
   }
}

void CGObjectMap::removeHandle(_CGhandle _handle) {
   CGObjectMapEntry *c = first;
   CGObjectMapEntry *p = NULL;
   while(NULL != c)
   {
      if(c->handle == _handle)
      {
         CGObjectMapEntry *n = c->next;
         if(NULL == p)
         {
            first = n;
         }
         else
         {
            p->next = n;
         }
         delete c;
         return;
      }
      p = c;
      c = c->next;
   }
}

