/*
 *    file: CGParameter.cpp
 *  author: Bastian Spiegel <bs@tkscript.de>
 *
 * changed: 21Nov2007, 26Dec2007
 *
 * license: MIT
 *
Copyright (c) 2007-2008 Bastian Spiegel, Carsten Busse

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

#include "tkcg.h"


_CGParameter::_CGParameter(void) {
   cg_parameter = NULL;
   texture_val = NULL;
}

_CGParameter::~_CGParameter() {
   destroy();
}

_CGhandle _CGParameter::getCgHandle(void) {
   return (_CGhandle) cg_parameter;
}

void _CGParameter::setCgHandle(_CGhandle _cgHandle) {
   cg_parameter = (CGparameter) _cgHandle;
   b_destroy = 0;
}

void _CGParameter::init(CGparameter _cgParameter) {
   destroy(); // Paranoia..
   cg_parameter = _cgParameter;
   b_destroy = 1;
}

void _CGParameter::destroy(void) {
   if(NULL != cg_parameter)
   {
      om_CGParameter.removeObject(this);
      if(b_destroy)
      {
         cgDestroyParameter(cg_parameter);
         Dhandlecgexception();
      }
      cg_parameter = NULL;
   }
   texture_val = NULL;
}

sSI _CGParameter::isReferenced(void) {
   CGbool ret = cgIsParameterReferenced(cg_parameter);
   if(Dhandlecgexception())
   {
      return (CG_TRUE==ret) ? 1 : 0;
   }
   else
   {
      return 0;
   }
}

void _CGParameter::setStateMatrix(sSI _a, sSI _b) {
   cgGLSetStateMatrixParameter(cg_parameter, (CGGLenum) _a, (CGGLenum) _b);
   Dhandlecgexception();
}

void _CGParameter::setTexture(YAC_Object *_texture) {
   if(YAC_VALID(_texture))
   {
      YAC_Value r;
      if(yac_host->yacEvalMethodByName(_texture, "getId", NULL, 0, &r))
      {
         if(YAC_TYPE_INT == r.type)
         {
            cgGLSetTextureParameter(cg_parameter, (GLuint) r.value.int_val);
            texture_val = _texture;
            Dhandlecgexception();
         }
         else
         {
            Draisetypemismatchexception("getId() method of <texture> object did not return an int value.");
         }
      }
      else
      {
         Draisetypemismatchexception("Failed to call getId() method of <texture> object.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid <texture> object.");
   }
}

void _CGParameter::setupSampler(YAC_Object *_texture) {
   if(YAC_VALID(_texture))
   {
      YAC_Value r;
      if(yac_host->yacEvalMethodByName(_texture, "getId", NULL, 0, &r))
      {
         if(YAC_TYPE_INT == r.type)
         {
            cgGLSetupSampler(cg_parameter, (GLuint) r.value.int_val);
            Dhandlecgexception();
         }
         else
         {
            Draisetypemismatchexception("getId() method of <texture> object did not return an int value.");
         }
      }
      else
      {
         Draisetypemismatchexception("Failed to call getId() method of <texture> object.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid <texture> object.");
   }
}

sSI _CGParameter::getTextureId(void) {
   sSI r = (sSI) cgGLGetTextureParameter(cg_parameter);
   Dhandlecgexception();
   return r;
}

YAC_Object *_CGParameter::getTexture(void) {
   sSI r = (sSI) cgGLGetTextureParameter(cg_parameter);
   if(Dhandlecgexception())
   {
      return texture_val;
   }
   else
   {
      texture_val = NULL;
      return NULL;
   }
}

void _CGParameter::enableTexture(void) {
   cgGLEnableTextureParameter(cg_parameter);
   Dhandlecgexception();
}

void _CGParameter::disableTexture(void) {
   cgGLDisableTextureParameter(cg_parameter);
   Dhandlecgexception();
}

void _CGParameter::setnf(YAC_Object *_o, sUI _num, const char *_oobMsg) {
   if(YAC_VALID(_o))
   {
      sUI eType = _o->yacArrayGetElementType();
      if( (YAC_TYPE_INT==eType) || (YAC_TYPE_FLOAT==eType) )
      {
         sF32 p[4];
         YAC_Value r;
         if(_o->yacArrayGetNumElements() >= _num)
         {
            sUI i;
            YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
            for(i=0; i<_num; i++)
            {
               _o->yacArrayGet(ctx, i, &r);
               if(YAC_TYPE_FLOAT != r.type)
               {
                  r.value.float_val = (sF32)r.value.int_val;
               }
               p[i] = r.value.float_val;
            }

            switch(_num)
            {
               case 1:
                  cgSetParameter1fv(cg_parameter, p);
                  Dhandlecgexception();
                  break;

               case 2:
                  cgSetParameter2fv(cg_parameter, p);
                  Dhandlecgexception();
                  break;

               case 3:
                  cgSetParameter3fv(cg_parameter, p);
                  Dhandlecgexception();
                  break;

               case 4:
                  cgSetParameter4fv(cg_parameter, p);
                  Dhandlecgexception();
                  break;
            }
         }
         else
         {
            Draisearrayoutofboundsexception(_oobMsg);
         }
      }
      else
      {
         Draisetypemismatchexception("Unsupported array element type.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid array object.");
   }
}

void _CGParameter::setni(YAC_Object *_o, sUI _num, const char *_oobMsg) {
   if(YAC_VALID(_o))
   {
      sUI eType = _o->yacArrayGetElementType();
      if( (YAC_TYPE_INT==eType) || (YAC_TYPE_FLOAT==eType) )
      {
         sSI p[4];
         YAC_Value r;
         if(_o->yacArrayGetNumElements() >= _num)
         {
            sUI i;
            YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
            for(i=0; i<_num; i++)
            {
               _o->yacArrayGet(ctx, i, &r);
               if(YAC_TYPE_INT != r.type)
               {
                  r.value.int_val = (sSI)r.value.float_val;
               }
               p[i] = r.value.int_val;
            }

            switch(_num)
            {
               case 1:
                  cgSetParameter1iv(cg_parameter, p);
                  Dhandlecgexception();
                  break;

               case 2:
                  cgSetParameter2iv(cg_parameter, p);
                  Dhandlecgexception();
                  break;

               case 3:
                  cgSetParameter3iv(cg_parameter, p);
                  Dhandlecgexception();
                  break;

               case 4:
                  cgSetParameter4iv(cg_parameter, p);
                  Dhandlecgexception();
                  break;
            }
         }
         else
         {
            Draisearrayoutofboundsexception(_oobMsg);
         }
      }
      else
      {
         Draisetypemismatchexception("Unsupported array element type.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid array object.");
   }
}

void _CGParameter::set1i(sSI _a) {
   cgSetParameter1i(cg_parameter, _a);
   Dhandlecgexception();
}

void _CGParameter::set1iv(YAC_Object *_o) {
   setni(_o, 1, "Array must have at least 1 element.");
}

void _CGParameter::set1f(sF32 _a) {
   cgSetParameter1f(cg_parameter, _a);
   Dhandlecgexception();
}

void _CGParameter::set1fv(YAC_Object *_o) {
   setnf(_o, 1, "Array must have at least 1 element.");
}

void _CGParameter::set2i(sSI _a, sSI _b) {
   cgSetParameter2i(cg_parameter, _a, _b);
   Dhandlecgexception();
}

void _CGParameter::set2iv(YAC_Object *_o) {
   setni(_o, 2, "Array must have at least 2 elements.");
}

void _CGParameter::set2f(sF32 _a, sF32 _b) {
   cgSetParameter2f(cg_parameter, _a, _b);
   Dhandlecgexception();
}

void _CGParameter::set2fv(YAC_Object *_o) {
   setnf(_o, 2, "Array must have at least 3 elements.");
}

void _CGParameter::set3i(sSI _a, sSI _b, sSI _c) {
   cgSetParameter3i(cg_parameter, _a, _b, _c);
   Dhandlecgexception();
}

void _CGParameter::set3iv(YAC_Object *_o) {
   setni(_o, 3, "Array must have at least 3 elements.");
}

void _CGParameter::set3f(sF32 _a, sF32 _b, sF32 _c) {
   cgSetParameter3f(cg_parameter, _a, _b, _c);
   Dhandlecgexception();
}

void _CGParameter::set3fv(YAC_Object *_o) {
   setnf(_o, 3, "Array must have at least 3 elements.");
}

void _CGParameter::set4i(sSI _a, sSI _b, sSI _c, sSI _d) {
   cgSetParameter4i(cg_parameter, _a, _b, _c, _d);
   Dhandlecgexception();
}

void _CGParameter::set4iv(YAC_Object *_o) {
   setni(_o, 4, "Array must have at least 4 elements.");
}

void _CGParameter::set4f(sF32 _a, sF32 _b, sF32 _c, sF32 _d) {
   cgSetParameter4f(cg_parameter, _a, _b, _c, _d);
   Dhandlecgexception();
}

void _CGParameter::set4fv(YAC_Object *_o) {
   setnf(_o, 4, "Array must have at least 4 elements.");
}

void _CGParameter::enableClientState(void) {
   cgGLEnableClientState(cg_parameter);
   Dhandlecgexception();
}

void _CGParameter::disableClientState(void) {
   cgGLDisableClientState(cg_parameter);
   Dhandlecgexception();
}

void _CGParameter::getMatrixf(YAC_Object *_faRet, bool _bRowLayout) {
   if(YAC_VALID(_faRet))
   {
      if(_faRet->yacArrayGetMaxElements() >= 16)
      {
         if(YAC_TYPE_FLOAT == _faRet->yacArrayGetElementType())
         {
            sF32 *fap;
            static sF32 fa[16];
            if(YAC_BCHK(_faRet, YAC_CLID_FLOATARRAY))
            {
               fap = (sF32*) _faRet->yacArrayGetPointer();
            }
            else
            {
               fap = fa;
            }
            if(_bRowLayout)
            {
               cgGLGetMatrixParameterfr(cg_parameter, fap);
            }
            else
            {
               cgGLGetMatrixParameterfc(cg_parameter, fap);
            }
            if(Dhandlecgexception())
            {
               if(fap != fa)
               {
                  sUI i;
                  YAC_Value r;
                  r.type = YAC_TYPE_FLOAT;
                  YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
                  for(i=0; i<16; i++)
                  {
                     r.value.float_val = fa[i];
                     _faRet->yacArraySet(ctx, i, &r);
                  }
               }
               //_faRet->yacArraySetNumElements(16);
            }
         }
         else
         {
            Draisetypemismatchexception("Expected float element type.");
         }
      }
      else
      {
         Draisearrayoutofboundsexception("Array must be able to hold at least 16 elements.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid Array object.");
   }
}

void _CGParameter::getMatrixd(YAC_Object *_noaRet, bool _bRowLayout) {
   if(YAC_VALID(_noaRet))
   {
      if(_noaRet->yacArrayGetMaxElements() >= 16)
      {
         if(YAC_TYPE_OBJECT == _noaRet->yacArrayGetElementType())
         {
            static sF64 da[16];
            if(_bRowLayout)
            {
               cgGLGetMatrixParameterdr(cg_parameter, da);
            }
            else
            {
               cgGLGetMatrixParameterdc(cg_parameter, da);
            }
            if(Dhandlecgexception())
            {
               sUI i;
               YAC_Value r;
               YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
               for(i=0; i<16; i++)
               {
                  _noaRet->yacArrayGet(ctx, i, &r); // Get number object
                  if(YAC_VALID(r.value.object_val))
                  {
                     // Set value of number object (probably Double)
                     r.value.object_val->yacValueOfF64(da[i]);
                  }
                  else
                  {
                     Draiseinvalidpointerexception("Invalid array element number object.");
                     return;
                  }
               }
               //_noaRet->yacArraySetNumElements(16);
            }
         }
         else
         {
            Draisetypemismatchexception("Expected (number-)object element type.");
         }
      }
      else
      {
         Draisearrayoutofboundsexception("Array must be able to hold at least 16 elements.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid Array object.");
   }
}

void _CGParameter::getMatrixfr(YAC_Object *_faRet) {
   getMatrixf(_faRet, true);
}

void _CGParameter::getMatrixfc(YAC_Object *_faRet) {
   getMatrixf(_faRet, false);
}

void _CGParameter::getMatrixdr(YAC_Object *_noaRet) {
   getMatrixd(_noaRet, true);
}

void _CGParameter::getMatrixdc(YAC_Object *_noaRet) {
   getMatrixd(_noaRet, false);
}

sSI _CGParameter::getTextureEnum(void) {
   sSI r = cgGLGetTextureEnum(cg_parameter);
   Dhandlecgexception();
   return r;
}

void _CGParameter::setMatrixf(YAC_Object *_fa, bool _bRowLayout) {
   if(YAC_VALID(_fa))
   {
      if(_fa->yacArrayGetNumElements() >= 16)
      {
         if(YAC_TYPE_FLOAT == _fa->yacArrayGetElementType())
         {
            sF32 *fap;
            static sF32 fa[16];
            if(YAC_BCHK(_fa, YAC_CLID_FLOATARRAY))
            {
               fap = (sF32*) _fa->yacArrayGetPointer();
            }
            else
            {
               sUI i;
               YAC_Value r;
               YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
               for(i=0; i<16; i++)
               {
                  _fa->yacArrayGet(ctx, i, &r);
                  fa[i] = r.value.float_val;
               }
               fap = fa;
            }
            if(_bRowLayout)
            {
               cgGLSetMatrixParameterfr(cg_parameter, fap);
               Dhandlecgexception();
            }
            else
            {
               cgGLSetMatrixParameterfc(cg_parameter, fap);
               Dhandlecgexception();
            }
         }
         else
         {
            Draisetypemismatchexception("Expected float element type.");
         }
      }
      else
      {
         Draisearrayoutofboundsexception("Array must have at least 16 elements.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid Array object.");
   }
}

void _CGParameter::setMatrixd(YAC_Object *_noa, bool _bRowLayout) {
   if(YAC_VALID(_noa))
   {
      if(_noa->yacArrayGetNumElements() >= 16)
      {
         if(YAC_TYPE_OBJECT == _noa->yacArrayGetElementType())
         {
            static sF64 da[16];
            sUI i;
            YAC_Value r;
            YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
            for(i=0; i<16; i++)
            {
               _noa->yacArrayGet(ctx, i, &r);
               if(YAC_VALID(r.value.object_val))
               {
                  r.value.object_val->yacScanF64(&da[i]);
               }
               else
               {
                  Draiseinvalidpointerexception("Invalid array element (number-)object.");
                  return;
               }
            }
            if(_bRowLayout)
            {
               cgGLSetMatrixParameterdr(cg_parameter, da);
               Dhandlecgexception();
            }
            else
            {
               cgGLSetMatrixParameterdc(cg_parameter, da);
               Dhandlecgexception();
            }
         }
         else
         {
            Draisetypemismatchexception("Expected (number-)object element type.");
         }
      }
      else
      {
         Draisearrayoutofboundsexception("Array must have at least 16 elements.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid Array object.");
   }
}

void _CGParameter::setMatrixfr(YAC_Object *_fa) {
   setMatrixf(_fa, true);
}

void _CGParameter::setMatrixfc(YAC_Object *_fa) {
   setMatrixf(_fa, false);
}

void _CGParameter::setMatrixdr(YAC_Object *_noa) {
   setMatrixd(_noa, true);
}

void _CGParameter::setMatrixdc(YAC_Object *_noa) {
   setMatrixd(_noa, false);
}

void _CGParameter::disconnect(void) {
   cgDisconnectParameter(cg_parameter);
   Dhandlecgexception();
}

void _CGParameter::getConnectedParameter(YAC_Value *_r) {

   _r->initNull();

   CGparameter cgParameter = cgGetConnectedParameter(cg_parameter);
   if(Dhandlecgexception())
   {
      om_CGParameter.findOrCreateByHandle(cgParameter, clid_CGParameter, _r);
   }

}

void _CGParameter::getConnectedToParameter(sSI _index, YAC_Value *_r) {
   _r->initNull();

   CGparameter cgParameter = cgGetConnectedToParameter(cg_parameter, _index);
   if(Dhandlecgexception())
   {
      om_CGParameter.findOrCreateByHandle(cgParameter, clid_CGParameter, _r);
   }
}

void _CGParameter::connectTo(YAC_Object *_cgParameter) {
   if(YAC_CHK(_cgParameter, clid_CGParameter))
   {
      cgConnectParameter(cg_parameter, ((_CGParameter*)_cgParameter)->cg_parameter);
      Dhandlecgexception();
   }
   else
   {
      Draiseinvalidpointerexception("Invalid CGParameter object parameter.");
   }
}

