/*
 *    file: CGProgram.cpp
 *  author: Bastian Spiegel <bs@tkscript.de>
 *
 * changed: 21Nov2007, 26Dec2007
 *
 * license: MIT
 *
Copyright (c) 2007-2008 Bastian Spiegel, Carsten Busse

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

#include "tkcg.h"


_CGProgram::_CGProgram(void) {
   cg_program = NULL;
}

_CGProgram::~_CGProgram() {
   destroy();
}

_CGhandle _CGProgram::getCgHandle(void) {
   return (_CGhandle) cg_program;
}

void _CGProgram::setCgHandle(_CGhandle _cgHandle) {
   cg_program = (CGprogram) _cgHandle;
   b_destroy = 0;
}

void _CGProgram::init(CGprogram _cgProgram) {
   destroy(); // Paranoia :)
   cg_program = _cgProgram;
   b_destroy = 1;
}

void _CGProgram::destroy(void) {
   if(NULL != cg_program)
   {
      om_CGProgram.removeObject(this);
      if(b_destroy)
      {
         cgDestroyProgram(cg_program);
         Dhandlecgexception();
         b_destroy = 0;
      }
      cg_program = NULL;
   }
}

void _CGProgram::bind(void) {
   cgGLBindProgram(cg_program);
   Dhandlecgexception();
}

void _CGProgram::load(void) {
   cgGLLoadProgram(cg_program);
   Dhandlecgexception();
}

void _CGProgram::getNamedParameter(YAC_String *_name, YAC_Value *_r) {
   YAC_Value r;

   if(YAC_CHK(_name, YAC_CLID_STRING))
   {
      YAC_String *name = (YAC_String *) _name;

      CGparameter cgParameter = cgGetNamedParameter(cg_program, (char*)name->chars);
      if(Dhandlecgexception())
      {
         om_CGParameter.findOrCreateByHandle(cgParameter, clid_CGParameter, _r);
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid <name> object.");
   }
}

void _CGProgram::enableProfiles(void) {
   cgGLEnableProgramProfiles(cg_program);
   Dhandlecgexception();
}

void _CGProgram::disableProfiles(void) {
   cgGLDisableProgramProfiles(cg_program);
   Dhandlecgexception();
}

sSI _CGProgram::getId(void) {
   sSI ret =(sSI) cgGLGetProgramID(cg_program);
   Dhandlecgexception();
   return ret;
}

sSI _CGProgram::isLoaded(void) {
   CGbool ret = cgGLIsProgramLoaded(cg_program);
   if(Dhandlecgexception())
   {
      return (CG_TRUE==ret) ? 1 : 0;
   }
   else
   {
      return 0;
   }
}

sSI _CGProgram::isCompiled(void) {
   CGbool ret = cgIsProgramCompiled(cg_program);
   if(Dhandlecgexception())
   {
      return (CG_TRUE==ret) ? 1 : 0;
   }
   else
   {
      return 0;
   }
}

void _CGProgram::copy(YAC_Value *_r) {
   CGprogram cgProgram = cgCopyProgram(cg_program);

   if(Dhandlecgexception())
   {
      _CGProgram *p = (_CGProgram *) yac_host->yacNewByID(clid_CGProgram);
      if(NULL != p)
      {
         p->init(cgProgram);
         _r->initObject(p, 1);
         om_CGProgram.addObject(p);
      }
      else
      {
         cgDestroyProgram(cgProgram);
         Draiseinvalidpointerexception("Failed to create CGProgram object.");
      }
   }
}

