/*
 *    file: tkcg.cpp
 *  author: Bastian Spiegel <bs@tkscript.de>
 *          Carsten Busse <carsten.busse@googlemail.com>
 *
 * changed: 21Nov2007, 26Dec2007
 *
 * license: MIT
 *
Copyright (c) 2007-2008 Bastian Spiegel, Carsten Busse

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/


#include "tkcg.h"

#include "ying_CG.h"

#include "ying_CG_CGContext.cpp"
#include "ying_CG_CGProfile.cpp"
#include "ying_CG_CGProgram.cpp"
#include "ying_CG_CGParameter.cpp"
#include "ying_CG_CGEffect.cpp"
#include "ying_CG_CGState.cpp"
#include "ying_CG_CGStateAssignment.cpp"
#include "ying_CG_CGTechnique.cpp"
#include "ying_CG_CGObject.cpp"
#include "ying_CG.cpp"

#include "../tksdl/tksdl_hooks.h"
static tksdl_hook_t hook_open;
static tksdl_hook_t hook_close;

YAC_Host *yac_host;

CGObjectMap om_CGParameter;
CGObjectMap om_CGProgram;
CGObjectMap om_CGState;
CGObjectMap om_CGStateAssignment;
CGObjectMap om_CGTechnique;
CGObjectMap om_CGEffect;
CGObjectMap om_CGContext;

static YAC_ExceptionId exid_Error; 

YAC_ExceptionId exid_InvalidPointer;
YAC_ExceptionId exid_TypeMismatch;
YAC_ExceptionId exid_ClassTypeMismatch;
YAC_ExceptionId exid_ArrayOutOfBounds;

YAC_ExceptionId exid_CGError; // Base exception
YAC_ExceptionId exid_CGCompilerError;
YAC_ExceptionId exid_CGInvalidParameterError;
YAC_ExceptionId exid_CGInvalidProfileError;
YAC_ExceptionId exid_CGProgramLoadError;
YAC_ExceptionId exid_CGProgramBindError;
YAC_ExceptionId exid_CGProgramNotLoadedError;
YAC_ExceptionId exid_CGUnsupportedGLExtensionError;
YAC_ExceptionId exid_CGInvalidValueTypeError;
YAC_ExceptionId exid_CGNotMatrixParamError;
YAC_ExceptionId exid_CGInvalidEnumerantError;
YAC_ExceptionId exid_CGNot4x4MatrixError;
YAC_ExceptionId exid_CGFileReadError;
YAC_ExceptionId exid_CGFileWriteError;
YAC_ExceptionId exid_CGNVParseError;
YAC_ExceptionId exid_CGMemoryAllocError;
YAC_ExceptionId exid_CGInvalidContextHandleError;
YAC_ExceptionId exid_CGInvalidProgramHandleError;
YAC_ExceptionId exid_CGInvalidParamHandleError;
YAC_ExceptionId exid_CGUnknownProfileError;
YAC_ExceptionId exid_CGVarArgError;
YAC_ExceptionId exid_CGInvalidDimensionError;
YAC_ExceptionId exid_CGArrayParamError;
YAC_ExceptionId exid_CGOutOfArrayBoundsError;
YAC_ExceptionId exid_CGConflictingTypesError;
YAC_ExceptionId exid_CGConflictingParameterTypesError;
YAC_ExceptionId exid_CGParameterIsNotSharedError;
YAC_ExceptionId exid_CGInvalidParameterVariabilityError;
YAC_ExceptionId exid_CGCannotDestroyParameterError;
YAC_ExceptionId exid_CGNotRootParameterError;
YAC_ExceptionId exid_CGParametersDoNotMatchError;
YAC_ExceptionId exid_CGIsNotProgramParameterError;
YAC_ExceptionId exid_CGInvalidParameterTypeError;
YAC_ExceptionId exid_CGParameterIsNotResizableArrayError;
YAC_ExceptionId exid_CGInvalidSizeError;
YAC_ExceptionId exid_CGBindCreatesCycleError;
YAC_ExceptionId exid_CGArrayTypesDoNotMatchError;
YAC_ExceptionId exid_CGArrayDimensionsDoNotMatchError;
YAC_ExceptionId exid_CGArrayHasWrongDimensionError;
YAC_ExceptionId exid_CGTypeIsNotDefinedInProgramError;
YAC_ExceptionId exid_CGInvalidEffectHandleError;
YAC_ExceptionId exid_CGInvalidStateHandleError;
YAC_ExceptionId exid_CGInvalidStateAssignmentHandleError;
YAC_ExceptionId exid_CGInvalidPassHandleError;
YAC_ExceptionId exid_CGInvalidAnnotationHandleError;
YAC_ExceptionId exid_CGInvalidTechniqueHandleError;
YAC_ExceptionId exid_CGStateAssignmentTypeMismatchError;
YAC_ExceptionId exid_CGInvalidFunctionHandleError;
YAC_ExceptionId exid_CGInvalidTechniqueError;
YAC_ExceptionId exid_CGInvalidPointerError;
YAC_ExceptionId exid_CGNotEnoughDataError;
YAC_ExceptionId exid_CGNonNumericParameterError;
YAC_ExceptionId exid_CGArraySizeMismatchError;
YAC_ExceptionId exid_CGCannotSetNonUniformParameterError;
YAC_ExceptionId exid_CGDuplicateNameError;


sBool HandleCGException(const char *_file, sSI _line) {
   // Return true if no exception was raised

   CGerror err = cgGetError();
   switch(err)
   {
      case CG_NO_ERROR:  
         return 1;

      default:
      {
         static char buf[256];
         sprintf(buf, "Unknown CG error (%d) occured.", err); 
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGError, 
                                     buf, _file, _line); 
      }
      return 0;
      
      case CG_COMPILER_ERROR: 
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGCompilerError, 
                                     "The compile returned an error", _file, _line); 
         return 0;
         
      case CG_INVALID_PARAMETER_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidParameterError, 
                                     "The parameter used is invalid.", _file, _line); 
         return 0;

      case CG_INVALID_PROFILE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidProfileError, 
                                     "The profile is not supported.", _file, _line); 
         return 0;
         
      case CG_PROGRAM_LOAD_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGProgramLoadError, 
                                     "The program could not load.", _file, _line); 
         return 0;

      case CG_PROGRAM_BIND_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGProgramBindError, 
                                     "The program could not bind.", _file, _line); 
         return 0;

      case CG_PROGRAM_NOT_LOADED_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGProgramNotLoadedError, 
                                     "The program must be loaded before this operation may be used.", _file, _line); 
         return 0;

      case CG_UNSUPPORTED_GL_EXTENSION_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGUnsupportedGLExtensionError, 
                                     "An unsupported GL extension was required to perform this operation.", _file, _line); 
         return 0;

      case CG_INVALID_VALUE_TYPE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidValueTypeError, 
                                     "An unknown value type was assigned to a parameter.", _file, _line); 
         return 0;

      case CG_NOT_MATRIX_PARAM_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGNotMatrixParamError, 
                                     "The parameter is not of matrix type.", _file, _line); 
         return 0;
         
      case CG_INVALID_ENUMERANT_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidEnumerantError, 
                                     "The enumerant parameter has an invalid value.", _file, _line); 
         return 0;

      case CG_NOT_4x4_MATRIX_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGNot4x4MatrixError,
                                     "The parameter must be a 4x4 matrix type.", _file, _line); 
         return 0;

      case CG_FILE_READ_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGFileReadError, 
                                     "The file could not be read.", _file, _line); 
         return 0;

      case CG_FILE_WRITE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGFileWriteError, 
                                     "The file could not be written.", _file, _line); 
         return 0;

      case CG_NVPARSE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGNVParseError,
                                     "nvparse could not successfully parse the output from the Cg compiler backend.", _file, _line); 
         return 0;

      case CG_MEMORY_ALLOC_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGMemoryAllocError, 
                                     "Memory allocation failed.", _file, _line); 
         return 0;

      case CG_INVALID_CONTEXT_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidContextHandleError, 
                                     "Invalid context handle.", _file, _line); 
         return 0;

      case CG_INVALID_PROGRAM_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidProgramHandleError, 
                                     "Invalid program handle.", _file, _line); 
         return 0;

      case CG_INVALID_PARAM_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidParamHandleError, 
                                     "Invalid parameter handle.", _file, _line); 
         return 0;

      case CG_UNKNOWN_PROFILE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGUnknownProfileError,
                                     "The specified profile is unknown.", _file, _line); 
         return 0;

      case CG_VAR_ARG_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGVarArgError,
                                     "The variable arguments were specified incorrectly.", _file, _line); 
         return 0;

      case CG_INVALID_DIMENSION_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidDimensionError,
                                     "The dimension value is invalid.", _file, _line); 
         return 0;

      case CG_ARRAY_PARAM_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGArrayParamError,
                                     "The parameter must be an array.", _file, _line); 
         return 0;

      case CG_OUT_OF_ARRAY_BOUNDS_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGOutOfArrayBoundsError, 
                                     "Index into the array is out of bounds.", _file, _line); 
         return 0;

      case CG_CONFLICTING_TYPES_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGConflictingTypesError, 
                                     "A type being added to the context conflicts with an existing type.", _file, _line); 
         return 0;

      case CG_CONFLICTING_PARAMETER_TYPES_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGConflictingParameterTypesError,
                                     "The parameters being bound have conflicting types.", _file, _line); 
         return 0;

      case CG_PARAMETER_IS_NOT_SHARED_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGParameterIsNotSharedError,
                                     "The parameter must be global.", _file, _line); 
         return 0;

      case CG_INVALID_PARAMETER_VARIABILITY_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidParameterVariabilityError, 
                                     "The parameter could not be changed to the given variability.", _file, _line); 
         return 0;

      case CG_CANNOT_DESTROY_PARAMETER_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGCannotDestroyParameterError,
                                     "Cannot destroy the parameter.  It is bound to other parameters or is not a root parameter.", _file, _line);
         return 0;

      case CG_NOT_ROOT_PARAMETER_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGNotRootParameterError, 
                                     "The parameter is not a root parameter.", _file, _line); 
         return 0;

      case CG_PARAMETERS_DO_NOT_MATCH_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGParametersDoNotMatchError, 
                                     "The two parameters being bound do not match.", _file, _line); 
         return 0;

      case CG_IS_NOT_PROGRAM_PARAMETER_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGIsNotProgramParameterError, 
                                     "The parameter is not a program parameter.", _file, _line); 
         return 0;

      case CG_INVALID_PARAMETER_TYPE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidParameterTypeError, 
                                     "The type of the parameter is invalid.", _file, _line); 
         return 0;

      case CG_PARAMETER_IS_NOT_RESIZABLE_ARRAY_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGParameterIsNotResizableArrayError,
                                     "The parameter must be a resizable array.", _file, _line); 
         return 0;

      case CG_INVALID_SIZE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidSizeError,
                                     "The size value is invalid.", _file, _line); 
         return 0;

      case CG_BIND_CREATES_CYCLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGBindCreatesCycleError,
                                     "Cannot bind the given parameters.  Binding will form a cycle.", _file, _line); 
         return 0;
         
      case CG_ARRAY_TYPES_DO_NOT_MATCH_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGArrayTypesDoNotMatchError,
                                     "Cannot bind the given parameters.  Array types do not match.", _file, _line); 
         return 0;

      case CG_ARRAY_DIMENSIONS_DO_NOT_MATCH_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGArrayDimensionsDoNotMatchError,
                                     "Cannot bind the given parameters. Array dimensions do not match.", _file, _line); 
         return 0;

      case CG_ARRAY_HAS_WRONG_DIMENSION_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGArrayHasWrongDimensionError,
                                     "The array has the wrong dimension.", _file, _line); 
         return 0;

      case CG_TYPE_IS_NOT_DEFINED_IN_PROGRAM_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGTypeIsNotDefinedInProgramError,
                                     "Connecting the parameters failed because The type of the source parameter is not defined within the given program or does not match the type with the same name in the program.", _file, _line); 
         return 0;

      case CG_INVALID_EFFECT_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidEffectHandleError,
                                     "Invalid effect handle.", _file, _line); 
         return 0;

      case CG_INVALID_STATE_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidStateHandleError, 
                                     "Invalid state handle.", _file, _line); 
         return 0;

      case CG_INVALID_STATE_ASSIGNMENT_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidStateAssignmentHandleError,
                                     "Invalid stateassignment handle.", _file, _line); 
         return 0;

      case CG_INVALID_PASS_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidPassHandleError, 
                                     "Invalid pass handle.", _file, _line); 
         return 0;

      case CG_INVALID_ANNOTATION_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidAnnotationHandleError,
                                     "Invalid annotation handle.", _file, _line); 
         return 0;

      case CG_INVALID_TECHNIQUE_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidTechniqueHandleError,
                                     "Invalid technique handle.", _file, _line); 
         return 0;

      case CG_STATE_ASSIGNMENT_TYPE_MISMATCH_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGStateAssignmentTypeMismatchError, 
                                     "Operation is not valid for this type of stateassignment.", _file, _line); 
         return 0;

      case CG_INVALID_FUNCTION_HANDLE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidFunctionHandleError,
                                     "Invalid function handle.", _file, _line); 
         return 0;

      case CG_INVALID_TECHNIQUE_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidTechniqueError, 
                                     "Technique did not pass validation.", _file, _line); 
         return 0;

      case CG_INVALID_POINTER_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGInvalidPointerError, 
                                     "The supplied pointer is NULL.", _file, _line); 
         return 0;

      case CG_NOT_ENOUGH_DATA_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGNotEnoughDataError,
                                     "Not enough data was provided.", _file, _line); 
         return 0;

      case CG_NON_NUMERIC_PARAMETER_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGNonNumericParameterError,
                                     "The parameter is not of a numeric type.", _file, _line); 
         return 0;
         
      case CG_ARRAY_SIZE_MISMATCH_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGArraySizeMismatchError,
                                     "The specified array sizes are not compatible with the given array.", _file, _line); 
         return 0;

      case CG_CANNOT_SET_NON_UNIFORM_PARAMETER_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGCannotSetNonUniformParameterError,
                                     "Cannot set the value of a non-uniform parameter.", _file, _line); 
         return 0;

      case CG_DUPLICATE_NAME_ERROR:
         yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGDuplicateNameError, 
                                     "This name is already in use.", _file, _line); 
         return 0;
   }
}


void _cgGLSetDebugMode(sSI _bDebug) {
   cgGLSetDebugMode(_bDebug ? CG_TRUE : CG_FALSE);
   Dhandlecgexception();
}

sSI _cgGetEnum(YAC_String *_name) {
   if(YAC_BCHK(_name, YAC_CLID_STRING))
   {
      CGenum e = cgGetEnum( (char*) ((YAC_String*)_name)->chars );
      Dhandlecgexception();
      return (sSI) e;
   }
   else
   {
      Draiseinvalidpointerexception("Invalid <name> String parameter.");
   }
   return 0;
}
 

void _cgCombinePrograms(YAC_Object *_cgPrograms, YAC_Value *_r) {
   
   _r->initNull();

   if(YAC_VALID(_cgPrograms))
   {
      if(YAC_TYPE_OBJECT == _cgPrograms->yacArrayGetElementType())
      {
         sUI num = _cgPrograms->yacArrayGetNumElements();
         if(num >= 2)
         {
            static CGprogram programs[32];
            sUI i;
            YAC_Value r;
            YAC_ContextHandle ctx = yac_host->yacContextGetDefault();
            for(i=0; i<num; i++)
            {
               _cgPrograms->yacArrayGet(ctx, i, &r);
               if(YAC_CHK(r.value.object_val, clid_CGProgram))
               {
                  programs[i] = ((_CGProgram *)r.value.object_val)->cg_program;
               }
               else
               {
                  Draiseinvalidpointerexception("Invalid array element, expected CGParameter object.");
                  return;
               }
            }
            CGprogram cgProgram = cgCombinePrograms(num, programs);
            if(Dhandlecgexception())
            {
               _CGProgram *p = (_CGProgram *) yac_host->yacNewByID(clid_CGProgram);
               if(NULL != p)
               {
                  p->init(cgProgram);
                  _r->initObject(p, 1);
               }
               else
               {
                  Draiseinvalidpointerexception("Failed to create CGProgram object.");
               }
            }
         }
         else
         {
            Draisecgexception("CGParameter array must have at least two elements.");
         }
      }
      else
      {
         Draisetypemismatchexception("Exception Object array element type.");
      }
   }
   else
   {
      Draiseinvalidpointerexception("Invalid <cgParameters> object parameter.");
   }
}

void tkcg_shader_hook_open(void) {
}

void tkcg_shader_hook_close(void) {
   om_CGParameter.deleteEntries();
   om_CGProgram.deleteEntries();
   om_CGState.deleteEntries();
   om_CGStateAssignment.deleteEntries();
   om_CGTechnique.deleteEntries();
   om_CGEffect.deleteEntries();
   om_CGContext.deleteEntries();
}

void YAC_CALL YAC_Init(YAC_Host *_host) {
	yac_host = _host;

	YAC_Init_CG(_host);

   // Query basic TKS exceptions
   exid_Error             = yac_host->yacExceptionGetIdByName("CriticalError"); // maybe make it Uncritical? hmm...
	exid_ClassTypeMismatch = yac_host->yacExceptionGetIdByName("ClassTypeMismatch");
	exid_InvalidPointer    = yac_host->yacExceptionGetIdByName("InvalidPointer");
	exid_TypeMismatch      = yac_host->yacExceptionGetIdByName("TypeMismatch");
	exid_ArrayOutOfBounds  = yac_host->yacExceptionGetIdByName("ArrayOutOfBounds");

   // Register CG exceptions
   exid_CGError = yac_host->yacExceptionRegister("CGError", exid_Error); // Base exception

   exid_CGCompilerError                     = yac_host->yacExceptionRegister("CGCompilerError", exid_CGError);
   exid_CGInvalidParameterError             = yac_host->yacExceptionRegister("CGInvalidParameterError", exid_CGError);
   exid_CGInvalidProfileError               = yac_host->yacExceptionRegister("CGInvalidProfileError", exid_CGError);
   exid_CGProgramLoadError                  = yac_host->yacExceptionRegister("CGProgramLoadError", exid_CGError);
   exid_CGProgramBindError                  = yac_host->yacExceptionRegister("CGProgramBindError", exid_CGError);
   exid_CGProgramNotLoadedError             = yac_host->yacExceptionRegister("CGProgramNotLoadedError", exid_CGError);
   exid_CGUnsupportedGLExtensionError       = yac_host->yacExceptionRegister("CGUnsupportedGLExtensionError", exid_CGError);
   exid_CGInvalidValueTypeError             = yac_host->yacExceptionRegister("CGInvalidValueTypeError", exid_CGError);
   exid_CGNotMatrixParamError               = yac_host->yacExceptionRegister("CGNotMatrixParamError", exid_CGError);
   exid_CGInvalidEnumerantError             = yac_host->yacExceptionRegister("CGInvalidEnumerantError", exid_CGError);
   exid_CGNot4x4MatrixError                 = yac_host->yacExceptionRegister("CGNot4x4MatrixError", exid_CGError);
   exid_CGFileReadError                     = yac_host->yacExceptionRegister("CGFileReadError", exid_CGError);
   exid_CGFileWriteError                    = yac_host->yacExceptionRegister("CGFileWriteError", exid_CGError);
   exid_CGNVParseError                      = yac_host->yacExceptionRegister("CGNVParseError", exid_CGError);
   exid_CGMemoryAllocError                  = yac_host->yacExceptionRegister("CGMemoryAllocError", exid_CGError);
   exid_CGInvalidContextHandleError         = yac_host->yacExceptionRegister("CGInvalidContextHandleError", exid_CGError);
   exid_CGInvalidProgramHandleError         = yac_host->yacExceptionRegister("CGInvalidProgramHandleError", exid_CGError);
   exid_CGInvalidParamHandleError           = yac_host->yacExceptionRegister("CGInvalidParamHandleError", exid_CGError);
   exid_CGUnknownProfileError               = yac_host->yacExceptionRegister("CGUnknownProfileError", exid_CGError);
   exid_CGVarArgError                       = yac_host->yacExceptionRegister("CGVarArgError", exid_CGError);
   exid_CGInvalidDimensionError             = yac_host->yacExceptionRegister("CGInvalidDimensionError", exid_CGError);
   exid_CGArrayParamError                   = yac_host->yacExceptionRegister("CGArrayParamError", exid_CGError);
   exid_CGOutOfArrayBoundsError             = yac_host->yacExceptionRegister("CGOutOfArrayBoundsError", exid_CGError);
   exid_CGConflictingTypesError             = yac_host->yacExceptionRegister("CGConflictingTypesError", exid_CGError);
   exid_CGConflictingParameterTypesError    = yac_host->yacExceptionRegister("CGConflictingParameterTypesError", exid_CGError);
   exid_CGParameterIsNotSharedError         = yac_host->yacExceptionRegister("CGParameterIsNotSharedError", exid_CGError);
   exid_CGInvalidParameterVariabilityError  = yac_host->yacExceptionRegister("CGInvalidParameterVariabilityError", exid_CGError);
   exid_CGCannotDestroyParameterError       = yac_host->yacExceptionRegister("CGCannotDestroyProgramError", exid_CGError);
   exid_CGNotRootParameterError             = yac_host->yacExceptionRegister("CGNotRootParameterError", exid_CGError);
   exid_CGParametersDoNotMatchError         = yac_host->yacExceptionRegister("CGParametersDoNotMatchError", exid_CGError);
   exid_CGIsNotProgramParameterError        = yac_host->yacExceptionRegister("CGIsNotProgramParameterError", exid_CGError);
   exid_CGInvalidParameterTypeError         = yac_host->yacExceptionRegister("CGInvalidParameterTypeError", exid_CGError);
   exid_CGParameterIsNotResizableArrayError = yac_host->yacExceptionRegister("CGParameterIsNotResizableArrayError", exid_CGError);
   exid_CGInvalidSizeError                  = yac_host->yacExceptionRegister("CGInvalidSizeError", exid_CGError);
   exid_CGBindCreatesCycleError             = yac_host->yacExceptionRegister("CGBindCreatesCycleError", exid_CGError);
   exid_CGArrayTypesDoNotMatchError         = yac_host->yacExceptionRegister("CGArrayTypesDoNotMatchError", exid_CGError);
   exid_CGArrayDimensionsDoNotMatchError    = yac_host->yacExceptionRegister("CGArrayDimensionsDoNotMatchError", exid_CGError);
   exid_CGArrayHasWrongDimensionError       = yac_host->yacExceptionRegister("CGArrayHasWrongDimensionError", exid_CGError);
   exid_CGTypeIsNotDefinedInProgramError    = yac_host->yacExceptionRegister("CGTypeIsNotDefinedInProgramError", exid_CGError);
   exid_CGInvalidEffectHandleError          = yac_host->yacExceptionRegister("CGInvalidEffectHandleError", exid_CGError);
   exid_CGInvalidStateHandleError           = yac_host->yacExceptionRegister("CGInvalidStateHandleError", exid_CGError);
   exid_CGInvalidStateAssignmentHandleError = yac_host->yacExceptionRegister("CGInvalidStateAssignmentHandleError", exid_CGError);
   exid_CGInvalidPassHandleError            = yac_host->yacExceptionRegister("CGInvalidPassHandleError", exid_CGError);
   exid_CGInvalidAnnotationHandleError      = yac_host->yacExceptionRegister("CGInvalidAnnotationHandleError", exid_CGError);
   exid_CGInvalidTechniqueHandleError       = yac_host->yacExceptionRegister("CGInvalidTechniqueHandleError", exid_CGError);
   exid_CGStateAssignmentTypeMismatchError  = yac_host->yacExceptionRegister("CGStateAssignmentTypeMismatchError", exid_CGError);
   exid_CGInvalidFunctionHandleError        = yac_host->yacExceptionRegister("CGInvalidFunctionHandleError", exid_CGError);
   exid_CGInvalidTechniqueError             = yac_host->yacExceptionRegister("CGInvalidTechniqueError", exid_CGError);
   exid_CGInvalidPointerError               = yac_host->yacExceptionRegister("CGInvalidPointerError", exid_CGError);
   exid_CGNotEnoughDataError                = yac_host->yacExceptionRegister("CGNotEnoughDataError", exid_CGError);
   exid_CGNonNumericParameterError          = yac_host->yacExceptionRegister("CGNonNumericParameterError", exid_CGError);
   exid_CGArraySizeMismatchError            = yac_host->yacExceptionRegister("CGArraySizeMismatchError", exid_CGError);
   exid_CGCannotSetNonUniformParameterError = yac_host->yacExceptionRegister("CGCannotSetNonUniformParameterError", exid_CGError);
   exid_CGDuplicateNameError                = yac_host->yacExceptionRegister("CGDuplicateNameError", exid_CGError);

   hook_open.fxn = &tkcg_shader_hook_open;
   hook_open.next = NULL;
   tksdl_hook_add(TKSDL_HOOK_NAME_OPEN, &hook_open);

   hook_close.fxn = &tkcg_shader_hook_close;
   hook_close.next = NULL;
   tksdl_hook_add(TKSDL_HOOK_NAME_CLOSE, &hook_close);
}

void YAC_CALL YAC_Exit(YAC_Host *_host) {
	YAC_Exit_CG(_host);
}

#include <yac_host.cpp>
