/*
 *    file: tkcg.h
 *  author: Bastian Spiegel <bs@tkscript.de>
 *
 * changed: 21Nov2007, 26Dec2007
 *
 * license: MIT
 *
Copyright (c) 2007-2008 Bastian Spiegel, Carsten Busse

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef __TKCG_H__
#define __TKCG_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define YAC_PRINTF
#define YAC_BIGSTRING
#include <yac.h>
extern YAC_Host *yac_host;

#include <math.h>
#ifdef YAC_WIN32
#include <windows.h>
//#include <ddraw.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#include "ying_CG.h"

#include "CGObject.h"
#include "CGProfile.h"
#include "CGParameter.h"
#include "CGProgram.h"
#include "CGState.h"
#include "CGStateAssignment.h"
#include "CGTechnique.h"
#include "CGEffect.h"
#include "CGContext.h"
#include "CGObjectMap.h"

YG("CG");

extern CGObjectMap om_CGParameter;
extern CGObjectMap om_CGProgram;
extern CGObjectMap om_CGState;
extern CGObjectMap om_CGStateAssignment;
extern CGObjectMap om_CGTechnique;
extern CGObjectMap om_CGEffect;
extern CGObjectMap om_CGContext;

extern YAC_ExceptionId exid_InvalidPointer;
extern YAC_ExceptionId exid_TypeMismatch;
extern YAC_ExceptionId exid_ClassTypeMismatch;
extern YAC_ExceptionId exid_ArrayOutOfBounds;

#define Draiseinvalidpointerexception(a) yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_InvalidPointer, (a), __FILE__, __LINE__)

#define Draisetypemismatchexception(a) yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_TypeMismatch, (a), __FILE__, __LINE__)

#define Draisearrayoutofboundsexception(a) yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_ArrayOutOfBounds, (a), __FILE__, __LINE__)

#define Draisecgexception(a) yac_host->yacExceptionRaise(yac_host->yacContextGetDefault(), exid_CGError, (a), __FILE__, __LINE__)


extern YAC_ExceptionId exid_CGError; // Base exception
extern YAC_ExceptionId exid_CGCompilerError;
extern YAC_ExceptionId exid_CGInvalidParameterError;
extern YAC_ExceptionId exid_CGInvalidProfileError;
extern YAC_ExceptionId exid_CGProgramLoadError;
extern YAC_ExceptionId exid_CGProgramBindError;
extern YAC_ExceptionId exid_CGProgramNotLoadedError;
extern YAC_ExceptionId exid_CGUnsupportedGLExtensionError;
extern YAC_ExceptionId exid_CGInvalidValueTypeError;
extern YAC_ExceptionId exid_CGNotMatrixParamError;
extern YAC_ExceptionId exid_CGInvalidEnumerantError;
extern YAC_ExceptionId exid_CGNot4x4MatrixError;
extern YAC_ExceptionId exid_CGFileReadError;
extern YAC_ExceptionId exid_CGFileWriteError;
extern YAC_ExceptionId exid_CGNVParseError;
extern YAC_ExceptionId exid_CGMemoryAllocError;
extern YAC_ExceptionId exid_CGInvalidContextHandleError;
extern YAC_ExceptionId exid_CGInvalidProgramHandleError;
extern YAC_ExceptionId exid_CGInvalidParamHandleError;
extern YAC_ExceptionId exid_CGUnknownProfileError;
extern YAC_ExceptionId exid_CGVarArgError;
extern YAC_ExceptionId exid_CGInvalidDimensionError;
extern YAC_ExceptionId exid_CGArrayParamError;
extern YAC_ExceptionId exid_CGOutOfArrayBoundsError;
extern YAC_ExceptionId exid_CGConflictingTypesError;
extern YAC_ExceptionId exid_CGConflictingParameterTypesError;
extern YAC_ExceptionId exid_CGParameterIsNotSharedError;
extern YAC_ExceptionId exid_CGInvalidParameterVariabilityError;
extern YAC_ExceptionId exid_CGCannotDestroyParameterError;
extern YAC_ExceptionId exid_CGNotRootParameterError;
extern YAC_ExceptionId exid_CGParametersDoNotMatchError;
extern YAC_ExceptionId exid_CGIsNotProgramParameterError;
extern YAC_ExceptionId exid_CGInvalidParameterTypeError;
extern YAC_ExceptionId exid_CGParameterIsNotResizableArrayError;
extern YAC_ExceptionId exid_CGInvalidSizeError;
extern YAC_ExceptionId exid_CGBindCreatesCycleError;
extern YAC_ExceptionId exid_CGArrayTypesDoNotMatchError;
extern YAC_ExceptionId exid_CGArrayDimensionsDoNotMatchError;
extern YAC_ExceptionId exid_CGArrayHasWrongDimensionError;
extern YAC_ExceptionId exid_CGTypeIsNotDefinedInProgramError;
extern YAC_ExceptionId exid_CGInvalidEffectHandleError;
extern YAC_ExceptionId exid_CGInvalidStateHandleError;
extern YAC_ExceptionId exid_CGInvalidStateAssignmentHandleError;
extern YAC_ExceptionId exid_CGInvalidPassHandleError;
extern YAC_ExceptionId exid_CGInvalidAnnotationHandleError;
extern YAC_ExceptionId exid_CGInvalidTechniqueHandleError;
extern YAC_ExceptionId exid_CGStateAssignmentTypeMismatchError;
extern YAC_ExceptionId exid_CGInvalidFunctionHandleError;
extern YAC_ExceptionId exid_CGInvalidTechniqueError;
extern YAC_ExceptionId exid_CGInvalidPointerError;
extern YAC_ExceptionId exid_CGNotEnoughDataError;
extern YAC_ExceptionId exid_CGNonNumericParameterError;
extern YAC_ExceptionId exid_CGArraySizeMismatchError;
extern YAC_ExceptionId exid_CGCannotSetNonUniformParameterError;
extern YAC_ExceptionId exid_CGDuplicateNameError;


#define Dhandlecgexception() HandleCGException(__FILE__, __LINE__)

extern sBool HandleCGException(const char *_file, sSI _line);


YF void YAC_CALL _cgGLSetDebugMode  (sSI _bDebug);
YF sSI  YAC_CALL _cgGetEnum         (YAC_String *_name);
YF void YAC_CALL _cgCombinePrograms (YAC_Object *_cgParameterArray, YAC_Value *_r);


#endif // __TKCG_H__
